#!/usr/bin/env python
#
# Copyright (C) 2009 Leonid Movshovich <event.riga@gmail.com>
#
# ktorrent2deluge is a free software.
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# deluge is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with deluge.    If not, write to:
# 	The Free Software Foundation, Inc.,
# 	51 Franklin Street, Fifth Floor
# 	Boston, MA    02110-1301, USA.
#
# 
#

from sys import argv, exit
import os
import os.path
from deluge.core.torrentmanager import TorrentState, TorrentManagerState
import cPickle
import shutil

USAGE = '''This script transforms ktorrent state to deluge state so if you want to change your 
torrent client you don't need to move all the beloved many-years-seeded torrents by hand :)
USAGE: %1s <ktorrent dir> <deluge conf dir>
<ktorrent dir> - /home/user/.kde/share/apps/ktorrent or any other folder wher torXXX files are stored
<deluge conf dir> - /home/user/.config/deluge or any other folder where you want deluge state to be created.
After this script to finish with transition process please run deluge once and then ktorrent2deluge_post.py script.
'''

def bdecode(s) :
    """
    bdecode the string 's'. The result is tuple: (<bdecoded value>, <size>)
    <bdecoded value> could be string, int, list or dict depending on what is bencoded
    <size> is whole code size, e.g. '1:a' will have size 3, and 'i12312e' - 7
    """
    fch = s[0]
    if fch.isdigit() :
        scidx = s.index(':')
        L = int(s[:scidx])
        scidx += 1
        return (s[scidx : scidx + L], scidx + L)
    elif fch == 'i' :
        size = s.index('e') + 1
        return (int(s[1 : size - 1]), size)
    elif fch == 'l' :
        res = []
        end = 1
        while s[end] != 'e' :
            next = bdecode(s[end:])
            res.append(next[0])
            end += next[1]
        return (res, end + 1)
    elif fch == 'd' :
        res = {}
        end = 1
        while s[end] != 'e' :
            key = bdecode(s[end:])
            end += key[1]
            value = bdecode(s[end:])
            end += value[1]
            res[key[0]] = value[0]
        return (res, end + 1)
    else :
        raise ValueError('Not a valid bencoded sequence')
        


class tor_info(object):
    """
    deluge-friendly presentation of torrent being download. Properties:
    tor_file - torrent file path
    location - folder to save content
    trackers - additional trackers
    nodownload - list of integers representing files which should not be downloaded. 
                 Members are indecies of files ordered by their relative pathes
    file_cnt - count of files and folders
    """
    __slots__ = ['tor_file', 'location', 'trackers', 'nodownload', 'file_cnt', 'uploaded']
    def __init__(self, **keyargs) :
        for k, v in keyargs.items() :
            if v is not None :
                self.__setattr__(k, v)

    def __getstate__(self) :
        result = {}
        for fld in self.__slots__ :
            result[fld] = self.__getattr__(fld)
        return result

    def __setstate__(self, state) :
        tor_info.__init__(self, **state)
        

def getfiles(torrentdata) :
    """
    extracts a list of files (with relative paths if any) from torrentdata
    torrentdata - bdecoded torrent meta info
    """
    info = torrentdata['info']
    if info.has_key('files') :
        result = [os.path.join(*f['path']) for f in 
                  info['files']]
        result.sort()
        return result
    else :
        return [info['name']]

def get_dnd_files(dir, base) :
    result = []
    for root, dirs, files in os.walk(dir) :
        relroot = root.replace(base, '')
        if len(relroot) > 0 :
            relroot = relroot[1:]
        result.extend([os.path.join(relroot, f[:-4]) for f in files])
    return result

def getuploaded(statspath) :
    lines = open(statspath, 'rb').readlines()
    try :
        upline = filter(lambda x: x.startswith('UPLOADED'), lines)[0]
    except IndexError:
        return 0
    else :
        return long(upline[upline.index('=') + 1:])

def create_info(tordir) :
    print tordir
    tor = os.path.join(tordir, 'torrent')
    tordata = bdecode(open(tor, 'rb').read())[0]
    torfiles = getfiles(tordata)
    files_cnt = len(torfiles)
    loc = os.path.join(tordir, 'cache')
    if os.path.islink(loc) :
        location = os.path.split(os.readlink(loc))[0]
    else : # it's a folder
        walker = os.walk(loc)
        try :
            t = walker.next()
            root = t[0]
            files = t[2]
            while len(files) == 0 :
                t = walker.next()
                root = t[0]
                files = t[2]
        except StopIteration :
            location = None
        else : 
            somefile = os.path.join(root, files[0])
            redundandchars = len(somefile) - somefile.index('cache/') - 6 #len('cache/')
            location = os.readlink(somefile)[:-redundandchars]
            
    dnd = os.path.join(tordir, 'dnd')
    if os.path.exists(dnd) :
        nodnld = []
        dndfiles = get_dnd_files(dnd, dnd)
        for dndf in dndfiles :
            nodnld.append(torfiles.index(dndf))
    else :
        nodnld = []

    trackers = [tordata['announce']]

    uploaded = getuploaded(os.path.join(tordir, 'stats'))
    return tor_info(tor_file=tor, location=location
                    , file_cnt=files_cnt, nodownload=nodnld
                    , trackers=trackers, uploaded=uploaded)

def ktorrent_export(base_dir) :
    """
    get ktorrent torrents in the form of list of tor_info instances
    """ 
    result = []
    tordirs = filter(lambda x: x.startswith('tor'), os.listdir(base_dir))
    for d in tordirs :
        result.append(create_info(os.path.join(base_dir, d)))
    return result


def transform_trackers(trackers) :
    return [{'url': t, 'tier': 0} for t in trackers]

def deluge_import(torrents, base_dir) :
    """
    take list of tor_info and import it to deluge under base_dir
    """
    mgrstate = TorrentManagerState()
    id = 0
    statedir = os.path.join(base_dir, 'state')
    if not os.path.exists(statedir) :
        os.mkdir(statedir)
    for info in torrents :
        id += 1
        strid = str(id)
        shutil.copy(info.tor_file, os.path.join(statedir, strid + '.torrent'))
        priorities = [1 for i in xrange(info.file_cnt)]
        for dnd in info.nodownload :
            priorities[dnd] = 0

#class TorrentState:
#    def __init__(self,
#            torrent_id=None,
#            filename=None,
#            total_uploaded=0,
#            trackers=None,
#            compact=False,
#            paused=False,
#            save_path=None,
#            max_connections=-1,
#            max_upload_slots=-1,
#            max_upload_speed=-1.0,
#            max_download_speed=-1.0,
#            prioritize_first_last=False,
#            file_priorities=None,
#            queue=None,
#            auto_managed=True,
#            is_finished=False,
#            stop_ratio=2.00,
#            stop_at_ratio=False,
#            remove_at_ratio=False
#        ):

        torstate \
            = TorrentState(torrent_id=strid, filename=strid + '.torrent'\
                               , trackers=transform_trackers(info.trackers), paused=True\
                               , save_path=info.location\
                               , file_priorities=priorities\
                               , queue=-1\
                               , total_uploaded=long(info.uploaded))

        mgrstate.torrents.append(torstate)
    trgfile = open(os.path.join(base_dir, 'state', 'torrents.state'), 'wb')
    try :
        cPickle.dump(mgrstate, trgfile)
    finally : 
        trgfile.close()
    


if len(argv) != 3 :
    print USAGE % argv[0]
    exit(1)

print "Here we go!"

deluge_import(ktorrent_export(argv[1]), argv[2])

print "Done!"
